#!/usr/bin/env python3
#
# Call with pytest. Requires XKB_CONFIG_ROOT to be set

import os
import subprocess
import sys
from pathlib import Path

import pytest


def _xkb_config_root() -> list[Path]:
    roots: list[Path] = []

    # XKB_CONFIG_EXTRA_PATH is not mandatory
    if path := os.getenv("XKB_CONFIG_EXTRA_PATH"):
        roots.append(Path(path))
        print(f"Using XKB_CONFIG_EXTRA_PATH={path}")

    if path := os.getenv("XKB_CONFIG_ROOT"):
        xkbpath = Path(path)
        assert (xkbpath / "symbols").exists(), f"{path} is not an XKB installation"
        roots.append(xkbpath)
        print(f"Using XKB_CONFIG_ROOT={path}")
    else:
        raise ValueError("Environment variable XKB_CONFIG_ROOT must be set")

    return roots


@pytest.fixture
def xkb_config_root():
    return _xkb_config_root()


def pytest_generate_tests(metafunc):
    # for any test_foo function with an argument named xkb_symbols,
    # make it a list of tuples in the form (us, dvorak)
    # The list is generated by scanning all xkb symbol files and extracting the
    # various sections
    if "xkb_symbols" in metafunc.fixturenames:
        xkb_symbols = []
        # This skips the *_vndr directories, they're harder to test
        for root in _xkb_config_root():
            for symbols_file in root.glob("symbols/*"):
                if symbols_file.is_dir():
                    continue
                with open(symbols_file) as f:
                    print(f"Found file {symbols_file}")
                    for line in f:
                        if not line.startswith('xkb_symbols "'):
                            continue
                        section = line.split('"')[1]
                        entry = (symbols_file.name, section)
                        if entry not in xkb_symbols:
                            xkb_symbols.append(entry)
        assert xkb_symbols
        metafunc.parametrize("xkb_symbols", xkb_symbols)


def test_xkbcomp(xkb_config_root, xkb_symbols):
    layout, variant = xkb_symbols
    keymap = f"""
xkb_keymap {{
        xkb_keycodes  {{ include "evdev+aliases(qwerty)" }};
        xkb_types     {{ include "complete"      }};
        xkb_compat    {{ include "complete"      }};
        xkb_symbols   {{ include "pc+{layout}({variant})"     }};
}};
"""
    print(keymap)

    args = [
        "xkbcomp",
        "-w0",  # reduce warning nose
        "-xkb",
        "-I",  # disable system includes
    ]
    args += [f"-I{p}" for p in xkb_config_root]
    args += [
        "-",  # from stdin
        "-",  # to stdout
    ]
    p = subprocess.run(args, input=keymap, encoding="utf-8", capture_output=True)
    if p.stderr:
        print(p.stderr, file=sys.stderr)
    if p.returncode != 0:
        print(p.stdout)
    if p.returncode < 0:
        print(f"xkbcomp exited with signal {-p.returncode}", file=sys.stderr)

    assert p.returncode == 0



# SPDX-License-Identifier: MIT

from __future__ import annotations

import functools
import itertools
import os
import re
from collections.abc import Generator
from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING, ClassVar, Optional, Union

if TYPE_CHECKING:
    import builtins

import pytest
import xkbcommon


@dataclass
class RMLVO:
    rules: str
    model: Optional[str]
    layout: Optional[str]
    variant: Optional[str]
    options: Optional[str]


@functools.total_ordering
@dataclass(frozen=True, order=False)
class Layout:
    PATTERN: ClassVar[re.Pattern[str]] = re.compile(
        r"(?P<layout>[^(]+)\((?P<variant>[^)]+)\)"
    )

    layout: str
    variant: str

    @classmethod
    def parse(cls, layout: str, variant: str = ""):
        # parse a layout(variant) string
        if match := cls.PATTERN.match(layout):
            assert not variant
            layout = match.group("layout")
            variant = match.group("variant")
        return cls(layout, variant)

    def __str__(self):
        if self.variant:
            return f"{self.layout}({self.variant})"
        else:
            return f"{self.layout}"

    def __lt__(self, other):
        """
        Custom compare function in order to deal with missing variant.
        """
        if not isinstance(other, self.__class__):
            return NotImplemented
        elif self.layout == other.layout:
            if self.variant == other.variant:
                return False
            # Handle missing variant
            elif self.variant and not other.variant:
                return True
            # Handle missing variant
            elif not self.variant and other.variant:
                return False
            else:
                return self.variant < other.variant
        else:
            return self.layout < other.layout

    @classmethod
    def read_file(cls, path: Path) -> Generator[tuple[Layout, Layout], None, None]:
        """Returns a list of two-layout tuples [(layout1, layout2), ...]"""

        with path.open("rt", encoding="utf-8") as fd:
            for line in fd:
                # Remove optional comment
                line = line.split("//")[0]
                # Split on whitespaces
                groups = tuple(line.split())
                length = len(groups)
                if length == 2:
                    l1 = Layout.parse(groups[0])
                    l2 = Layout.parse(groups[1])
                elif length == 4:
                    l1 = Layout.parse(groups[0], groups[1])
                    l2 = Layout.parse(groups[2], groups[3])
                else:
                    raise ValueError(f"Invalid line: {line}")
                yield (l1, l2)


def pytest_generate_tests(metafunc: pytest.Metafunc):
    if "mapping" in metafunc.fixturenames:
        if files := metafunc.config.getoption("layout_compat_config"):
            # Read all files
            mappings = tuple(
                itertools.chain.from_iterable(map(Layout.read_file, files)),
            )
        else:
            mappings = ()
        metafunc.parametrize("mapping", mappings)


@pytest.fixture(scope="session")
def xkb_base():
    """Get the xkeyboard-config directory from the environment."""
    roots: list[Path] = []

    # XKB_CONFIG_EXTRA_PATH is not mandatory
    if path := os.environ.get("XKB_CONFIG_EXTRA_PATH"):
        roots.append(Path(path))

    if path := os.environ.get("XKB_CONFIG_ROOT"):
        roots.append(Path(path))
    else:
        raise ValueError("XKB_CONFIG_ROOT environment variable is not defined")

    return roots


def compile_keymap(
    xkb_base: list[Path],
    rules: str,
    ls: tuple[Union[Layout, builtins.ellipsis], ...],
    layout: Layout,
) -> tuple[str, str, str]:
    lsʹ: tuple[Layout, ...] = tuple(layout if x is Ellipsis else x for x in ls)
    alias_layout = ",".join(l.layout for l in lsʹ)
    alias_variant = ",".join(l.variant for l in lsʹ)
    km = xkbcommon.ForeignKeymap(
        xkb_base=xkb_base, rules=rules, layout=alias_layout, variant=alias_variant
    )
    return alias_layout, alias_variant, km.as_string()


COMPONENT_NAME_PATTERN = re.compile(
    r"""
    (?P<component>xkb_(?:keycodes|types|compatibility|symbols))
    \s+
    "[^"]*"
    """,
    re.VERBOSE,
)


def drop_component_name(match: re.Match[str]) -> str:
    return match.group("component")


@pytest.mark.parametrize("rules", ("base", "evdev"))
def test_compat_layout(
    xkb_base: list[Path], rules: str, mapping: tuple[Layout, Layout]
):
    alias = mapping[0]
    target = mapping[1]
    us = Layout("us", "")
    fr = Layout("fr", "")
    es = Layout("es", "")
    configs = (
        (...,),
        (us, ...),
        (..., us),
        (us, fr, ...),
        (us, fr, es, ...),
    )
    for ls in configs:
        # Compile alias
        alias_layout, alias_variant, alias_string = compile_keymap(
            xkb_base, rules, ls, alias
        )
        assert alias_string != "", (rules, alias_layout, alias_variant)

        # Compile target
        target_layout, target_variant, target_string = compile_keymap(
            xkb_base, rules, ls, target
        )
        assert target_string != "", (rules, target_layout, target_variant)

        # [HACK] Fix keycodes aliases
        if alias.layout == "de" and target.layout != "de":
            # libxkbcommon ≤ 1.11
            alias_string = alias_string.replace(
                "<LatZ>         = <AD06>", "<LatY>         = <AD06>"
            )
            alias_string = alias_string.replace(
                "<LatY>         = <AB01>", "<LatZ>         = <AB01>"
            )
            # libxkbcommon ≥ 1.12
            alias_string = alias_string.replace("<LatZ> = <AD06>", "<LatZ> = <AB01>")
            alias_string = alias_string.replace("<LatY> = <AB01>", "<LatY> = <AD06>")

        # [HACK] Discard components names
        alias_string = COMPONENT_NAME_PATTERN.sub(drop_component_name, alias_string)
        target_string = COMPONENT_NAME_PATTERN.sub(drop_component_name, target_string)

        # Keymap obtained using alias should be the same as if using its target directly
        assert alias_string == target_string, (
            rules,
            alias_layout,
            alias_variant,
            target_layout,
            target_variant,
        )



# SPDX-License-Identifier: MIT

"""
Rules parser
"""

from __future__ import annotations

import dataclasses
import math
import re
import textwrap
from dataclasses import dataclass
from enum import IntEnum, StrEnum, auto, unique
from functools import partial
from typing import ClassVar, Iterable, Self


def natural_sort_key(s, _nsre=re.compile(r"([0-9]+)")):
    return tuple(
        int(chunk) if chunk.isdigit() else (chunk.lower(), chunk)
        for chunk in _nsre.split(s)
    )


@unique
class Version(IntEnum):
    V2 = 2
    "Compatible with X11"
    V3 = 3

    @classmethod
    def parse(cls, raw: str) -> Self:
        for v in cls:
            if str(v) == raw or f"v{v}" == raw:
                return v
        raise ValueError(raw)


@unique
class MLVO(StrEnum):
    Model = auto()
    Layout = auto()
    Variant = auto()
    Option = auto()

    @classmethod
    def parse(cls, raw: str) -> Self:
        for c in cls:
            if c == raw:
                return c
        raise ValueError(raw)


class LayoutIndex(int):
    def numeric_indices(self, max: int) -> Iterable[int]:
        yield self


@unique
class LayoutRange(StrEnum):
    Single = auto()
    First = auto()
    Later = auto()
    Any = auto()

    @classmethod
    def parse(cls, raw: str) -> Self | LayoutIndex:
        try:
            return LayoutIndex(int(raw, base=10))
        except ValueError:
            pass
        for i in cls:
            if i == raw:
                return i
        raise ValueError()

    def numeric_indices(self, max: int) -> Iterable[int]:
        match self:
            case self.Single:
                yield 0
            case self.First:
                yield 0
                yield 1
            case self.Later:
                yield from range(2, max + 1)
            case self.Any:
                yield from range(0, max + 1)
            case _:
                raise ValueError(self)


@unique
class KcCGST(StrEnum):
    Keycodes = auto()
    Compat = auto()
    Geometry = auto()
    Symbols = auto()
    Types = auto()

    @classmethod
    def parse(cls, raw: str) -> Self:
        for c in cls:
            if c == raw:
                return c
        raise ValueError(raw)


@dataclass(frozen=True)
class Rule:
    """
    A rule associates a MLVO matcher to a section and its value.
    """

    RULE_PATTERN: ClassVar[re.Pattern[str]] = re.compile(
        r"""
        ^
        (?P<mlvo>(?:\s+[^!\s=]+)+)
        \s+=\s+
        (?P<section>\S+)
        \s*$
        """,
        re.VERBOSE,
    )

    SECTION_INDICES: ClassVar[re.Pattern[str]] = re.compile(
        r"(?P<index>\[%i\])|(?P<modifier>:%i)"
    )

    mlvo: tuple[str]
    section: str

    @classmethod
    def parse(cls, raw: str) -> Self:
        if (m := cls.RULE_PATTERN.match(raw)) is None:
            raise ValueError(raw)
        mlvo = tuple(m.group("mlvo").split())
        return cls(mlvo=mlvo, section=m.group("section"))

    @staticmethod
    def _to_numeric_index(k: int, m: re.Match[str]) -> str:
        if k <= 0:
            return ""
        elif m.group("index"):
            return f"[{k}]"
        else:
            return f":{k}"

    def to_numeric_index(self, k: int) -> Self:
        section = self.SECTION_INDICES.sub(
            partial(self._to_numeric_index, k), self.section
        )
        return dataclasses.replace(self, section=section)


@dataclass
class RulesSet:
    """
    A rules set: group of rules with same MLVO and section type.
    """

    RULES_HEADER_PATTERN: ClassVar[re.Pattern[str]] = re.compile(
        r"""
        ^!
        (?P<mlvo>(?:\s+(?:model|layout|variant|option)(?:\[(?:\d+|single|first|later|any)\])?)+)
        \s+=\s+
        (?P<kccgst>\w+)
        \s*$
        """,
        re.VERBOSE,
    )

    MLVO_PATTERN: ClassVar[re.Pattern[str]] = re.compile(
        r"(?P<component>\w+)\[(?P<range>\w+)\]"
    )

    TAB_SIZE: ClassVar[int] = 8
    HEADER_START: ClassVar[str] = "! "
    LEADING_SPACES_COUNT: ClassVar[int] = len(HEADER_START)
    LEADING_SPACES: ClassVar[str] = LEADING_SPACES_COUNT * " "

    mlvo: tuple[MLVO, ...]
    layout_range: LayoutRange | LayoutIndex
    kccgst: KcCGST
    rules: tuple[Rule, ...]

    @classmethod
    def parse_header(
        cls, raw: str
    ) -> tuple[tuple[MLVO, ...], LayoutRange | int, KcCGST] | None:
        if (m := cls.RULES_HEADER_PATTERN.match(raw)) is None:
            return None

        mlvo, layout_index = cls.parse_mvlo(m.group("mlvo"))
        kccgst = KcCGST.parse(m.group("kccgst"))
        return mlvo, layout_index, kccgst

    @classmethod
    def parse_mvlo(cls, raw: str) -> tuple[tuple[MLVO], LayoutRange | LayoutIndex]:
        mlvo: list[MLVO] = []
        layout_range: LayoutRange | LayoutIndex | None = None
        for c in raw.split():
            # TODO: check layout range consistency
            if m := cls.MLVO_PATTERN.match(c):
                component = MLVO.parse(m.group("component"))
                layout_range = LayoutRange.parse(m.group("range"))
            else:
                component = MLVO.parse(c)
            mlvo.append(component)
        if layout_range is None:
            layout_range = LayoutRange.Single
        return tuple(mlvo), layout_range

    def to_numeric_index(self, k: int) -> Self:
        rules = list(r.to_numeric_index(k) for r in self.rules)
        return dataclasses.replace(self, layout_range=k, rules=rules)

    def to_numeric_indices(self, max: int) -> Iterable[Self]:
        yield from map(self.to_numeric_index, self.layout_range.numeric_indices(max))

    def serialize_mlvo(self) -> Iterable[str]:
        for c in self.mlvo:
            if (c is MLVO.Layout or c is MLVO.Variant) and (self.layout_range != 0):
                yield f"{c}[{self.layout_range}]"
            else:
                yield c

    def serialize(self, version: Version, debug: bool = False) -> Iterable[str]:
        if version is Version.V2:
            for k, rs in enumerate(self.to_numeric_indices(4)):
                if k > 0:
                    yield "\n"
                yield from rs._serialize(debug=debug)
        else:
            yield from self._serialize(debug=debug)

    def _serialize(self, debug: bool) -> Iterable[str]:
        # Pretty-printing: Find the maximal length of each component.
        mlvo = tuple(self.serialize_mlvo())
        mlvo = (self.HEADER_START + mlvo[0],) + mlvo[1:]
        sizes = [len(c) + 1 for c in mlvo]
        groups = [False for _ in mlvo]
        for rule in self.rules:
            for k, c in enumerate(rule.mlvo):
                if has_group := Group.is_group_name(c):
                    groups[k] = True
                if k == 0:
                    xtra_space = self.LEADING_SPACES_COUNT - (1 if has_group else 0)
                else:
                    xtra_space = 1 if has_group else 0
                sizes[k] = max(sizes[k], len(c) + 1 + xtra_space)

        # Update sizes and print header
        for k, (c, s, g) in enumerate(zip(mlvo, sizes, groups)):
            s = sizes[k] = math.ceil(s / self.TAB_SIZE) * self.TAB_SIZE
            yield self.append_tabs(" " + c if g and k > 0 else c, s)
        yield "=\t"
        yield self.kccgst
        yield "\n"

        # Print rules
        def prepend_space(c: str, k: int):
            if Group.is_group_name(c):
                if k == 0:
                    return self.LEADING_SPACES[:-1] + c
                else:
                    return c
            elif k == 0:
                return self.LEADING_SPACES + c
            elif groups[k]:
                return " " + c
            else:
                return c

        for rule in self.rules:
            if debug:
                yield f"  // "
                yield rule
                yield "\n"
            yield from (
                self.append_tabs(prepend_space(c, k), s)
                for k, (c, s) in enumerate(zip(rule.mlvo, sizes))
            )
            yield "=\t"
            yield rule.section
            yield "\n"

    @classmethod
    def append_tabs(cls, text: str, target_length: int) -> str:
        """
        Appends tabs to reach the target string length.
        """
        length = len(text)
        if length > target_length:
            return text
        else:
            count = math.ceil((target_length - length) / cls.TAB_SIZE)
            return text + count * "\t"


@dataclass
class Include:
    """
    Include statement in rules files
    """

    INCLUDE_PATTERN: ClassVar[re.Pattern[str]] = re.compile(
        r"""^!\s+include\s+(?P<file>[^\s"]+)\s*$"""
    )

    file: str

    @classmethod
    def parse(cls, line: str) -> Self | None:
        if m := cls.INCLUDE_PATTERN.match(line):
            file = m.group("file")
            return cls(file)
        else:
            return None

    def serialize(self, *args, **kwargs) -> Iterable[str]:
        yield f'! include "{self.file}\n"'


@dataclass
class Group:
    """
    Group of items used in rules
    """

    GROUP_HEADER_PATTERN: ClassVar[re.Pattern[str]] = re.compile(
        r"""
        ^!\s+
        (?P<name>\$(?:\w|-)+)
        \s+=
        (?P<members>(?:\s+(?:\w|-)+)+)
        \s*$
        """,
        re.VERBOSE,
    )

    name: str
    "Groups name starts with $"
    members: set[str]
    component: MLVO | None = None
    description: str = ""

    @classmethod
    def parse(cls, line: str) -> Self | None:
        if m := cls.GROUP_HEADER_PATTERN.match(line):
            name = m.group("name")
            members = m.group("members").split()
            return cls(name=name, members=set(members))
        else:
            return None

    def _serialize(self) -> str:
        return f"! {self.name} = {' '.join(sorted(self.members, key=natural_sort_key))}"

    def serialize(self, *args, **kwargs) -> Iterable[str]:
        disabled = not self.members
        group = self._serialize()
        eq_index = group.index("=") + 2
        group = " \\\n".join(
            textwrap.wrap(group, width=78, subsequent_indent=eq_index * " ")
        )
        if disabled:
            group = textwrap.indent(group, "//")
        if self.description:
            description = textwrap.indent(self.description, "// ")
            group = f"{description}\n{group}"
        yield group
        yield "\n"

    @staticmethod
    def is_group_name(s: str) -> bool:
        return s.startswith("$")


@dataclass
class Comment:
    """
    A generic comment
    """

    text: str = ""

    def __bool__(self) -> bool:
        return bool(self.text)

    def __add__(self, other):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return dataclasses.replace(self, text=self._add(other.text))

    def __iadd__(self, other):
        if isinstance(other, str):
            self.text = self._add(other)
        elif isinstance(other, self.__class__):
            self.text = self._add(other.text)
        else:
            return NotImplemented
        return self

    def _add(self, text: str) -> str:
        return f"{self.text}\n{text}"

    @classmethod
    def parse(cls, line: str) -> tuple[str, Self] | None:
        parts = line.split("//", maxsplit=1)
        if len(parts) > 1:
            return (parts[0], cls(parts[1]))
        else:
            return None

    def serialize(self, *args, **kwargs) -> Iterable[str]:
        yield textwrap.indent(self.text + "\n", "//", lambda _: True)


@dataclass
class BlankLine:
    def serialize(self, *args, **kwargs) -> Iterable[str]:
        yield "\n"


@dataclass
class RulesFile:
    """
    A rules file
    """

    @classmethod
    def parse(
        cls, raw: Iterable[str]
    ) -> Iterable[Include | Group | RulesSet | Comment | BlankLine]:
        rules_set: RulesSet | None = None
        pending_comment: Comment | None = None
        pending_line: str = ""
        pending_line_number: int | None = None
        for line_number, raw_line in enumerate(raw, start=1):
            # Handle comment
            parts = raw_line.split("//", maxsplit=1)
            line, *rest = parts
            if rest:
                comment: Comment | None = Comment(rest[0].rstrip())
            else:
                comment = None
            # Handle EOL escape
            parts = line.split("\\")
            if (length := len(parts)) > 1:
                # Only valid if actually at the end of the line
                if length > 2 or (parts[-1] and parts[-1] != "\n"):
                    raise ValueError(line_number, parts, raw_line)
                else:
                    if pending_line_number is None:
                        pending_line_number = line_number
                    pending_line += raw_line[:-2]
                    continue

            # Complete pending line
            line = pending_line + line
            pending_line = ""

            # Handle line with only whitespaces or comment.
            if not line.lstrip():
                if comment is None:
                    # Yield then reset pending comment on blank line
                    if pending_comment is not None:
                        yield pending_comment
                        pending_comment = None
                    if rules_set is None:
                        yield BlankLine()
                elif pending_comment is None:
                    pending_comment = comment
                else:
                    # Gather comments
                    pending_comment += comment
                continue

            # Rules set header
            if (rs := RulesSet.parse_header(line)) is not None:
                if rules_set is not None:
                    yield rules_set
                mlvo, layout_range, kccgst = rs
                rules_set = RulesSet(
                    mlvo=mlvo, layout_range=layout_range, kccgst=kccgst, rules=[]
                )
            # Group
            elif (group := Group.parse(line)) is not None:
                if rules_set is not None:
                    yield rules_set
                    rules_set = None
                if pending_comment is not None:
                    group.description = textwrap.dedent(pending_comment.text)
                yield group
            # Include
            elif (inc := Include.parse(line)) is not None:
                if rules_set is not None:
                    yield rules_set
                    rules_set = None
                yield inc
            # Rules
            elif rules_set is not None:
                if comment is not None:
                    if comment is not None:
                        # Current comment has priority
                        pending_comment = comment
                try:
                    rule = Rule.parse(
                        # mlvo=mlvo,
                        # layout_range=rules_set.layout_range,
                        # section=rules_set.section,
                        raw=line,
                    )
                except ValueError:
                    print(f"[ERROR] Invalid line {line_number}: {raw_line}")
                    raise
                # TODO use comment for rule?
                rules_set.rules.append(rule)
            # Unsupported: error
            else:
                line_ref = (
                    str(line_number)
                    if pending_line_number is None
                    else f"{pending_line_number}-{line_number}"
                )
                raise ValueError(f"Cannot parse line {line_ref}: {line}")
            pending_line_number = None
            pending_line = ""
            pending_comment = None
        if pending_line:
            raise ValueError(pending_line)
        if rules_set is not None:
            yield rules_set
        if pending_comment is not None:
            yield pending_comment

    @classmethod
    def serialize(
        cls,
        items: Iterable[Include | Group | RulesSet | Comment | BlankLine],
        version: Version,
        debug: bool = False,
    ) -> Iterable[str]:
        last: Include | Group | RulesSet | Comment | BlankLine = BlankLine()
        for i in items:
            if isinstance(i, BlankLine) and isinstance(last, BlankLine):
                # Merge blank lines
                continue
            elif isinstance(i, BlankLine) or isinstance(last, BlankLine):
                # Explicit blank line
                pass
            elif not (isinstance(i, Group) and isinstance(last, Group)):
                # No explicit blank line: add one to separate 2 entries, except
                # if they are both groups, so that related groups are kept together.
                yield "\n"
            last = i
            yield from i.serialize(version=version, debug=debug)

    @classmethod
    def render(
        cls, rules: Iterable[str], version: Version, debug: bool = False
    ) -> Iterable[str]:
        yield from cls.serialize(cls.parse(rules), version=version, debug=debug)



#!/usr/bin/env python3
#
# Call with pytest. Requires XKB_CONFIG_ROOT to be set

import os
import xml.etree.ElementTree as ET
from pathlib import Path
from typing import Iterator

import pytest


def _xkb_config_root() -> list[Path]:
    roots: list[Path] = []

    # XKB_CONFIG_EXTRA_PATH is not mandatory
    if path := os.getenv("XKB_CONFIG_EXTRA_PATH"):
        roots.append(Path(path))
        print(f"Using XKB_CONFIG_EXTRA_PATH={path}")

    if path := os.getenv("XKB_CONFIG_ROOT"):
        xkbpath = Path(path)
        assert (xkbpath / "symbols").exists(), f"{path} is not an XKB installation"
        roots.append(xkbpath)
        print(f"Using XKB_CONFIG_ROOT={path}")
    else:
        raise ValueError("Environment variable XKB_CONFIG_ROOT must be set")

    return roots


@pytest.fixture
def xkb_config_root():
    return _xkb_config_root()


def iterate_layouts_variants(rules_xml):
    """
    Return an iterator of type (layout, variant) for each element in the XML
    file.
    """
    tree = ET.parse(rules_xml)
    root = tree.getroot()
    for layout in root.iter("layout"):
        yield layout, None

        for variant in layout.iter("variant"):
            yield layout, variant


def iterate_config_items(rules_xml):
    """
    Return an iterator of configItem elements
    """
    tree = ET.parse(rules_xml)
    root = tree.getroot()
    return root.iter("configItem")


def get_registry_files() -> Iterator[Path]:
    """
    List all registry files in <XKB_ROOT>/rules/*.xml and ensure unique name.
    """
    found: set[Path] = set()
    for xkb_root in _xkb_config_root():
        for path in xkb_root.glob("rules/*.xml"):
            if (file := path.relative_to(xkb_root)) in found:
                continue
            else:
                yield path
                found.add(file)


def pytest_generate_tests(metafunc):
    # For any test_foo function with an argument named rules_xml,
    # make it the list of <XKB_ROOT>/rules/*.xml files.
    # Use only the first XKB root.
    if "rules_xml" in metafunc.fixturenames:
        rules_xml = list(get_registry_files())
        assert rules_xml
        metafunc.parametrize("rules_xml", rules_xml)
    # for any test_foo function with an argument named layout,
    # make it a Layout wrapper class for all layout(variant) combinations
    elif "layout" in metafunc.fixturenames:
        rules_xml = list(get_registry_files())
        assert rules_xml
        layouts = []
        for f in rules_xml:
            for l, v in iterate_layouts_variants(f):
                layouts.append(Layout(f, l, v))
        metafunc.parametrize("layout", layouts)
    elif "config_item" in metafunc.fixturenames:
        rules_xml = list(get_registry_files())
        assert rules_xml
        config_items = []
        for f in rules_xml:
            for item in iterate_config_items(f):
                item = ConfigItem.from_elem(item)
                item.rulesfile = f
                config_items.append(item)
        metafunc.parametrize("config_item", config_items)


class Layout:
    """
    Wrapper class for layout/variants - both ConfigItems are available but
    the properties automatically pick the variant (if it exists) or the
    layout otherwise.
    """

    def __init__(self, rulesfile, layout, variant=None):
        self.rulesfile = rulesfile
        self.layout = ConfigItem.from_elem(layout)
        self.variant = ConfigItem.from_elem(variant) if variant is not None else None
        if variant is not None:
            self.name = f"{self.layout.name}({self.variant.name})"
        else:
            self.name = f"{self.layout.name}"

    @property
    def iso3166(self):
        if self.variant and self.variant.iso3166 is not None:
            return self.variant.iso3166 or []
        # inherit from parent
        return self.layout.iso3166 or []

    @property
    def iso639(self):
        if self.variant and self.variant.iso639 is not None:
            return self.variant.iso639 or []
        # inherit from parent
        return self.layout.iso639 or []

    @property
    def popularity(self):
        return self.variant.popularity if self.variant else self.layout.popularity

    @property
    def shortDescription(self):
        if self.variant and self.variant.shortDescription:
            return self.variant.shortDescription
        return self.layout.shortDescription


def prettyxml(element):
    return ET.tostring(element).decode("utf-8")


class ConfigItem:
    def __init__(self, name, shortDescription=None, description=None):
        self.name = name
        self.shortDescription = shortDescription
        self.description = description
        self.iso639 = None
        self.iso3166 = None
        self.popularity = None

    @classmethod
    def _fetch_subelement(cls, parent, name):
        sub_element = parent.findall(name)
        if sub_element is not None and len(sub_element) == 1:
            return sub_element[0]
        else:
            return None

    @classmethod
    def _fetch_subelement_text(cls, parent, name):
        sub_element = parent.findall(name)
        return [e.text for e in sub_element]

    @classmethod
    def _fetch_text(cls, parent, name):
        sub_element = cls._fetch_subelement(parent, name)
        if sub_element is None:
            return None
        return sub_element.text

    @classmethod
    def from_elem(cls, elem):
        try:
            ci_element = (
                elem
                if elem.tag == "configItem"
                else cls._fetch_subelement(elem, "configItem")
            )
            name = cls._fetch_text(ci_element, "name")
            assert name is not None
            # shortDescription and description are optional
            sdesc = cls._fetch_text(ci_element, "shortDescription")
            desc = cls._fetch_text(ci_element, "description")
            ci = ConfigItem(name, sdesc, desc)
            ci.popularity = ci_element.attrib.get("popularity")

            langlist = cls._fetch_subelement(ci_element, "languageList")
            if langlist is not None:
                ci.iso639 = cls._fetch_subelement_text(langlist, "iso639Id")

            countrylist = cls._fetch_subelement(ci_element, "countryList")
            if countrylist is not None:
                ci.iso3166 = cls._fetch_subelement_text(countrylist, "iso3166Id")

            return ci
        except AssertionError as e:
            endl = "\n"  # f{} cannot contain backslashes
            e.args = (f"\nFor element {prettyxml(elem)}\n{endl.join(e.args)}",)
            raise


def test_duplicate_layouts(rules_xml):
    tree = ET.parse(rules_xml)
    root = tree.getroot()
    layouts = {}
    for layout in root.iter("layout"):
        ci = ConfigItem.from_elem(layout)
        assert ci.name not in layouts, f"Duplicate layout {ci.name}"
        layouts[ci.name] = True

        variants = {}
        for variant in layout.iter("variant"):
            vci = ConfigItem.from_elem(variant)
            assert vci.name not in variants, (
                f"{rules_xml}: duplicate variant {ci.name}({vci.name}):\n{prettyxml(variant)}"
            )
            variants[vci.name] = True


def test_duplicate_models(rules_xml):
    tree = ET.parse(rules_xml)
    root = tree.getroot()
    models = {}
    for model in root.iter("model"):
        ci = ConfigItem.from_elem(model)
        assert ci.name not in models, f"Duplicate model {ci.name}"
        models[ci.name] = True


def test_exotic(config_item):
    """All items in extras should be marked exotic"""
    if config_item.rulesfile.stem.endswith("extras"):
        assert config_item.popularity == "exotic", (
            f"{config_item.rulesfile}: item {config_item.name} does not have popularity exotic"
        )
    else:
        assert config_item.popularity != "exotic", (
            f"{config_item.rulesfile}: item {config_item.name} has popularity exotic"
        )


def test_short_description(layout):
    assert layout.shortDescription, (
        f"{layout.rulesfile}: layout {layout.name} missing shortDescription"
    )


def test_iso3166(layout):
    """Typically layouts should specify at least one country code"""
    pycountry = pytest.importorskip("pycountry")
    country_codes = [c.alpha_2 for c in pycountry.countries]
    expected_without_country = [
        "ancient",  # from before current-day countries
        "apl",
        "bqn",  # programming
        "brai",  # Braille not specific to any country
        "custom",
        "epo",  # Esperanto not native to any country
        "trans",  # international
    ]

    for code in layout.iso3166:
        assert code in country_codes, (
            f'{layout.rulesfile}: unknown country code "{code}" in {layout.name}'
        )

    assert layout.iso3166 or layout.layout.name in expected_without_country, (
        f"{layout.rulesfile}: layout {layout.name} has no countries associated"
    )


def test_iso639(layout):
    """Typically layouts should should specify at least one language code"""
    pycountry = pytest.importorskip("pycountry")

    # A list of languages not in pycountry, so we need to special-case them
    special_langs = [
        "ber",  # Berber languages (collective), https://iso639-3.sil.org/code/ber
        "btb",  # Beti (Cameroon), https://iso639-3.sil.org/code/btb
        "fox",  # Formosan languages (collective), https://iso639-3.sil.org/code/fox
        "phi",  # Philippine languages (collective), https://iso639-3.sil.org/code/phi
        "ovd",  # Elfdalian, https://iso639-3.sil.org/code/ovd
    ]
    language_codes = [c.alpha_3 for c in pycountry.languages] + special_langs
    expected_without_language = ["brai", "custom", "trans"]

    for code in layout.iso639:
        try:
            assert code in language_codes, (
                f'{layout.rulesfile}: unknown language code "{code}" in {layout.name}'
            )
        except AssertionError as e:
            # needs pycountry 23.12.7, this code can be removed when our CI has pycountry new enough
            if code in ["pzh", "szy", "uon"]:
                pytest.skip("pycountry too old, skipping")
            else:
                raise e

    assert layout.iso639 or layout.layout.name in expected_without_language, (
        f"{layout.rulesfile}: layout {layout.name} has no languages associated"
    )



# SPDX-License-Identifier: MIT

import os
import sys
from pathlib import Path

import pytest

tests_dir = Path(__file__).parent.resolve()
sys.path.insert(0, str(tests_dir))

try:
    import xdist  # noqa: F401

    # Otherwise we get unknown hook 'pytest_xdist_auto_num_workers'
    def pytest_xdist_auto_num_workers(config):
        return os.getenv("FDO_CI_CONCURRENT", None)

except ImportError:
    pass


def pytest_addoption(parser: pytest.Parser):
    parser.addoption(
        "--layout-compat-config",
        action="append",
        default=[],
        type=Path,
        help="List of layout compatibility files",
    )



# SPDX-License-Identifier: MIT

from __future__ import annotations

import os
from collections.abc import Iterable
from ctypes import (
    CFUNCTYPE,
    POINTER,
    Structure,
    _Pointer,
    byref,
    c_char,
    c_char_p,
    c_int,
    c_size_t,
    c_uint32,
    c_void_p,
    cdll,
    create_string_buffer,
    string_at,
)
from ctypes.util import find_library
from dataclasses import dataclass
from enum import Enum, IntFlag
from functools import partial, reduce
from pathlib import Path
from typing import TYPE_CHECKING, NamedTuple, Optional, Sequence, TypeAlias

###############################################################################
# Types
###############################################################################


class allocated_c_char_p(c_char_p):
    """
    This class is used in place of c_char_p when we need to free it manually.
    Python would convert c_char_p to a bytes string and we would not be able to free it.
    """

    pass


class xkb_context(Structure):
    pass


class xkb_rule_names(Structure):
    _fields_ = [
        ("rules", POINTER(c_char)),
        ("model", POINTER(c_char)),
        ("layout", POINTER(c_char)),
        ("variant", POINTER(c_char)),
        ("options", POINTER(c_char)),
    ]


class xkb_keymap(Structure):
    pass


class xkb_state(Structure):
    pass


class xkb_key_direction(Enum):
    XKB_KEY_UP = c_int(0)
    XKB_KEY_DOWN = c_int(1)


# [HACK] Typing ctypes correctly is difficult. The following works, but there
#        could be another better way.
if TYPE_CHECKING:
    xkb_context_p: TypeAlias = _Pointer[xkb_context]
    xkb_rule_names_p: TypeAlias = _Pointer[xkb_rule_names]
    xkb_keymap_p: TypeAlias = _Pointer[xkb_keymap]
    xkb_state_p = _Pointer[xkb_state]
else:
    xkb_context_p = POINTER(xkb_context)
    xkb_rule_names_p = POINTER(xkb_rule_names)
    xkb_keymap_p = POINTER(xkb_keymap)
    xkb_state_p = POINTER(xkb_state)
xkb_context_flags = c_int
xkb_keymap_compile_flags = c_int
xkb_keymap_format = c_int
xkb_keycode_t = c_uint32
xkb_keysym_t = c_uint32
xkb_mod_index_t = c_uint32
xkb_led_index_t = c_uint32
xkb_level_index_t = c_uint32
xkb_layout_index_t = c_uint32
xkb_state_component = c_int
xkb_consumed_mode = c_int
xkb_keysym_flags = c_int

xkb_keymap_key_iter_t = CFUNCTYPE(None, xkb_keymap_p, xkb_keycode_t, c_void_p)


###############################################################################
# Constants
###############################################################################


XKB_CONTEXT_NO_DEFAULT_INCLUDES = 1 << 0
XKB_CONTEXT_NO_ENVIRONMENT_NAMES = 1 << 1
XKB_KEYCODE_INVALID = 0xFFFFFFFF
XKB_MOD_INVALID = 0xFFFFFFFF
XKB_STATE_MODS_EFFECTIVE = 1 << 3
XKB_CONSUMED_MODE_XKB = 0
XKB_KEYMAP_FORMAT_TEXT_V1 = 1
XKB_KEYSYM_NO_FLAGS = 0


class ModifierMask(IntFlag):
    """Built-in standard definitions of modifiers masks"""

    Shift = 1 << 0
    Lock = 1 << 1
    Control = 1 << 2
    Mod1 = 1 << 3
    Mod2 = 1 << 4
    Mod3 = 1 << 5
    Mod4 = 1 << 6
    Mod5 = 1 << 7


NoModifier = ModifierMask(0)
Shift = ModifierMask.Shift
Lock = ModifierMask.Lock
Control = ModifierMask.Control
Mod1 = ModifierMask.Mod1
Mod2 = ModifierMask.Mod2
Mod3 = ModifierMask.Mod3
Mod4 = ModifierMask.Mod4
Mod5 = ModifierMask.Mod5


###############################################################################
# Binding to libxkbcommon
###############################################################################


xkbcommon_path = os.environ.get("XKBCOMMON_LIB_PATH")

if xkbcommon_path:
    xkbcommon_path = str(Path(xkbcommon_path).resolve())
    xkbcommon = cdll.LoadLibrary(xkbcommon_path)
else:
    xkbcommon_path = find_library("xkbcommon")
    if xkbcommon_path:
        xkbcommon = cdll.LoadLibrary(xkbcommon_path)
    else:
        raise OSError("Cannot load libxbcommon")

xkbcommon.xkb_keysym_from_name.argtypes = [c_char_p, xkb_keysym_flags]
xkbcommon.xkb_keysym_from_name.restype = xkb_keysym_t

xkbcommon.xkb_context_new.argtypes = [xkb_context_flags]
xkbcommon.xkb_context_new.restype = POINTER(xkb_context)

xkbcommon.xkb_keymap_new_from_names.argtypes = [
    POINTER(xkb_context),
    POINTER(xkb_rule_names),
    xkb_keymap_compile_flags,
]
xkbcommon.xkb_keymap_new_from_names.restype = POINTER(xkb_keymap)

xkbcommon.xkb_keymap_key_by_name.argtypes = [POINTER(xkb_keymap), c_char_p]
xkbcommon.xkb_keymap_key_by_name.restype = xkb_keycode_t

xkbcommon.xkb_keymap_get_as_string.argtypes = [POINTER(xkb_keymap), xkb_keymap_format]
# Note: should be c_char_p; see comment in the definiton of allocated_c_char_p.
xkbcommon.xkb_keymap_get_as_string.restype = allocated_c_char_p

xkbcommon.xkb_keymap_key_for_each.argtypes = [
    xkb_keymap_p,
    xkb_keymap_key_iter_t,
    c_void_p,
]
xkbcommon.xkb_keymap_key_for_each.restype = None

xkbcommon.xkb_keymap_num_layouts_for_key.argtypes = [xkb_keymap_p, xkb_keycode_t]
xkbcommon.xkb_keymap_num_layouts_for_key.restype = xkb_layout_index_t

xkbcommon.xkb_keymap_num_levels_for_key.argtypes = [
    xkb_keymap_p,
    xkb_keycode_t,
    xkb_layout_index_t,
]
xkbcommon.xkb_keymap_num_levels_for_key.restype = xkb_level_index_t

xkbcommon.xkb_keymap_key_get_syms_by_level.argtypes = [
    xkb_keymap_p,
    xkb_keycode_t,
    xkb_layout_index_t,
    xkb_level_index_t,
    POINTER(POINTER(xkb_keysym_t)),
]
xkbcommon.xkb_keymap_key_get_syms_by_level.restype = c_int

xkbcommon.xkb_state_new.argtypes = [POINTER(xkb_keymap)]
xkbcommon.xkb_state_new.restype = POINTER(xkb_state)

xkbcommon.xkb_state_get_keymap.argtypes = [POINTER(xkb_state)]
xkbcommon.xkb_state_get_keymap.restype = POINTER(xkb_keymap)

xkbcommon.xkb_state_key_get_one_sym.argtypes = [POINTER(xkb_state), xkb_keycode_t]
xkbcommon.xkb_state_key_get_one_sym.restype = xkb_keysym_t

xkbcommon.xkb_keymap_led_get_name.argtypes = [POINTER(xkb_keymap), xkb_led_index_t]
xkbcommon.xkb_keymap_led_get_name.restype = c_char_p

xkbcommon.xkb_state_key_get_layout.argtypes = [POINTER(xkb_state), xkb_keycode_t]
xkbcommon.xkb_state_key_get_layout.restype = xkb_layout_index_t

xkbcommon.xkb_state_key_get_level.argtypes = [
    POINTER(xkb_state),
    xkb_keycode_t,
    xkb_layout_index_t,
]
xkbcommon.xkb_state_key_get_level.restype = xkb_level_index_t

xkbcommon.xkb_keymap_num_mods.argtypes = [POINTER(xkb_keymap)]
xkbcommon.xkb_keymap_num_mods.restype = xkb_mod_index_t

xkbcommon.xkb_keymap_mod_get_name.argtypes = [POINTER(xkb_keymap), xkb_mod_index_t]
xkbcommon.xkb_keymap_mod_get_name.restype = c_char_p

xkbcommon.xkb_keymap_mod_get_index.argtypes = [POINTER(xkb_keymap), c_char_p]
xkbcommon.xkb_keymap_mod_get_index.restype = xkb_mod_index_t

xkbcommon.xkb_state_mod_index_is_active.argtypes = [
    POINTER(xkb_state),
    xkb_mod_index_t,
    xkb_state_component,
]
xkbcommon.xkb_state_mod_index_is_active.restype = c_int

xkbcommon.xkb_state_mod_index_is_consumed2.argtypes = [
    POINTER(xkb_state),
    xkb_keycode_t,
    xkb_mod_index_t,
    xkb_consumed_mode,
]
xkbcommon.xkb_state_mod_index_is_consumed2.restype = c_int

xkbcommon.xkb_keymap_num_leds.argtypes = [POINTER(xkb_keymap)]
xkbcommon.xkb_keymap_num_leds.restype = xkb_led_index_t

xkbcommon.xkb_state_led_index_is_active.argtypes = [POINTER(xkb_state), xkb_led_index_t]
xkbcommon.xkb_state_led_index_is_active.restype = c_int

if libc_path := find_library("c"):
    libc = cdll.LoadLibrary(libc_path)
else:
    raise ValueError("Cannot import libc")


def load_keymap(
    xkb_config_roots: Sequence[Path],
    rules=None,
    model=None,
    layout=None,
    variant=None,
    options=None,
) -> xkb_keymap_p:
    # Create context
    context = xkbcommon.xkb_context_new(
        XKB_CONTEXT_NO_DEFAULT_INCLUDES | XKB_CONTEXT_NO_ENVIRONMENT_NAMES
    )
    if not context:
        raise ValueError("Couldn't create xkb context")
    for xkb_config_root in xkb_config_roots:
        raw_path = create_string_buffer(str(xkb_config_root).encode("utf-8"))
        xkbcommon.xkb_context_include_path_append(context, raw_path)
    rmlvo = xkb_rule_names(
        rules=create_string_buffer(rules.encode("utf-8")) if rules else None,
        model=create_string_buffer(model.encode("utf-8")) if model else None,
        layout=create_string_buffer(layout.encode("utf-8")) if layout else None,
        variant=create_string_buffer(variant.encode("utf-8")) if variant else None,
        options=create_string_buffer(options.encode("utf-8")) if options else None,
    )
    # Load keymap
    keymap = xkbcommon.xkb_keymap_new_from_names(context, byref(rmlvo), 0)
    if not keymap:
        raise ValueError(
            f"Failed to compile RMLVO: {rules=}, {model=}, {layout=}, "
            f"{variant=}, {options=}"
        )

    xkbcommon.xkb_context_unref(context)
    return keymap


def unref_keymap(keymap: xkb_keymap_p) -> None:
    xkbcommon.xkb_keymap_unref(keymap)


def keymap_as_string(keymap: xkb_keymap_p) -> str:
    p = xkbcommon.xkb_keymap_get_as_string(keymap, XKB_KEYMAP_FORMAT_TEXT_V1)
    s = string_at(p).decode("utf-8")
    libc.free(p)
    return s


def xkb_keymap_key_for_each(keymap: xkb_keymap_p, f, x):
    xkbcommon.xkb_keymap_key_for_each(keymap, xkb_keymap_key_iter_t(f), x)


def xkb_keymap_num_layouts_for_key(keymap: xkb_keymap_p, keycode: int) -> int:
    return xkbcommon.xkb_keymap_num_layouts_for_key(keymap, keycode)


def xkb_keymap_num_levels_for_key(
    keymap: xkb_keymap_p, keycode: int, layout: int
) -> int:
    return xkbcommon.xkb_keymap_num_levels_for_key(keymap, keycode, layout)


def xkb_keymap_key_get_syms_by_level(
    keymap: xkb_keymap_p, keycode: int, layout: int, level: int
) -> tuple[int, ...]:
    keysyms_array_t = POINTER(xkb_keysym_t)
    keysyms_array = keysyms_array_t()
    count = xkbcommon.xkb_keymap_key_get_syms_by_level(
        keymap, keycode, layout, level, byref(keysyms_array)
    )
    return tuple(keysyms_array[k] for k in range(0, count))


def new_state(keymap: xkb_keymap_p) -> xkb_state_p:
    state = xkbcommon.xkb_state_new(keymap)
    if not state:
        raise ValueError("Cannot create state")
    return state


def init_state(
    xkeyboard_config_paths: Sequence[Path],
    rules=None,
    model=None,
    layout=None,
    variant=None,
    options=None,
) -> xkb_state_p:
    keymap = load_keymap(xkeyboard_config_paths, rules, model, layout, variant, options)
    return new_state(keymap)


def unref_state(state: xkb_state_p) -> None:
    xkbcommon.xkb_state_unref(state)


def xkb_keymap_mod_get_index(keymap: xkb_keymap_p, name: str) -> str:
    return xkbcommon.xkb_keymap_mod_get_index(keymap, name.encode("utf-8"))


def xkb_keymap_mod_get_name(keymap: xkb_keymap_p, mod: int) -> str:
    return xkbcommon.xkb_keymap_mod_get_name(keymap, mod).decode("utf-8")


def xkb_keymap_get_mod_mask_from_names(
    keymap: xkb_keymap_p, names: Iterable[str]
) -> ModifierMask:
    mods = filter(
        lambda m: m != XKB_MOD_INVALID,
        (xkb_keymap_mod_get_index(keymap, name) for name in names),
    )
    return reduce(lambda acc, m: acc | ModifierMask(1 << m), mods, ModifierMask(0))


def xkb_keymap_get_usual_mod_mapping(
    keymap: xkb_keymap_p, mask: ModifierMask
) -> ModifierMask:
    mods = (
        "Alt" if Mod1 in mask else "",
        "Meta" if Mod1 in mask else "",
        "NumLock" if Mod2 in mask else "",
        "LevelFive" if Mod3 in mask else "",
        "Super" if Mod4 in mask else "",
        "LevelThree" if Mod5 in mask else "",
    )
    return mask | xkb_keymap_get_mod_mask_from_names(keymap, mods)


def xkb_keymap_led_get_name(keymap: xkb_keymap_p, led: int) -> str:
    return xkbcommon.xkb_keymap_led_get_name(keymap, led).decode("utf-8")


def xkb_keysym_from_name(name: str) -> int:
    return xkbcommon.xkb_keysym_from_name(name.encode("utf-8"), XKB_KEYSYM_NO_FLAGS)


def xkb_keysym_get_name(keysym: int) -> str:
    buf_len = 90
    buf = create_string_buffer(buf_len)
    n = xkbcommon.xkb_keysym_get_name(keysym, buf, c_size_t(buf_len))
    if n > 0:
        return buf.value.decode("utf-8")
    elif n >= buf_len:
        raise ValueError(f"Truncated: expected {buf_len}, got: {n + 1}.")
    else:
        raise ValueError(f"Unsupported keysym: {keysym}")


def xkb_state_key_get_utf8(state: xkb_state_p, key: int) -> str:
    buf_len = 8
    buf = create_string_buffer(buf_len)
    n = xkbcommon.xkb_state_key_get_utf8(state, key, buf, c_size_t(buf_len))
    if n >= buf_len:
        raise ValueError(f"Truncated: expected {buf_len}, got: {n + 1}.")
    else:
        return buf.value.decode("utf-8")


###############################################################################
# Process key events
###############################################################################


class Result(NamedTuple):
    keycode: int
    layout: int
    keysym: str
    unicode: str
    level: int
    active_mods: ModifierMask
    consumed_mods: ModifierMask
    leds: tuple[str, ...]

    @property
    def group(self) -> int:
        """Alias for Result.layout"""
        return self.layout


def process_key_event(
    state: xkb_state_p, key: str, direction: xkb_key_direction
) -> Result:
    keymap: xkb_keymap_p = xkbcommon.xkb_state_get_keymap(state)
    keycode: int = xkbcommon.xkb_keymap_key_by_name(
        keymap, create_string_buffer(key.encode("utf-8"))
    )
    if keycode == XKB_KEYCODE_INVALID:
        raise ValueError(f"Unsupported key name: {key}")

    # Modifiers
    mods_count = xkbcommon.xkb_keymap_num_mods(keymap)
    active_mods = reduce(
        (lambda acc, m: acc | ModifierMask(1 << m)),
        filter(
            lambda m: xkbcommon.xkb_state_mod_index_is_active(
                state, m, XKB_STATE_MODS_EFFECTIVE
            ),
            range(0, mods_count),
        ),
        ModifierMask(0),
    )
    consumed_mods = reduce(
        (lambda acc, m: acc | ModifierMask(1 << m)),
        filter(
            lambda m: xkbcommon.xkb_state_mod_index_is_active(
                state, m, XKB_STATE_MODS_EFFECTIVE
            )
            and xkbcommon.xkb_state_mod_index_is_consumed2(
                state, keycode, m, XKB_CONSUMED_MODE_XKB
            ),
            range(0, mods_count),
        ),
        ModifierMask(0),
    )

    # Level
    layout: int = xkbcommon.xkb_state_key_get_layout(state, keycode)
    level: int = xkbcommon.xkb_state_key_get_level(state, keycode, layout)

    # Keysyms
    # [TODO] multiple keysyms
    # keysyms = (xkb_keysym_t * 2)()
    # keysyms_count = xkb_state_key_get_syms(state, keycode, byref(keysyms))
    keysym = xkbcommon.xkb_state_key_get_one_sym(state, keycode)
    keysym_str = xkb_keysym_get_name(keysym)

    # [TODO] Compose?
    # Unicode
    unicode = xkb_state_key_get_utf8(state, keycode)

    # LEDs
    leds_count = xkbcommon.xkb_keymap_num_leds(keymap)
    leds = tuple(
        xkb_keymap_led_get_name(keymap, led)
        for led in range(0, leds_count)
        if xkbcommon.xkb_state_led_index_is_active(state, led)
    )

    # Update state
    # [TODO] use return value?
    xkbcommon.xkb_state_update_key(state, keycode, direction.value)

    # Return result
    return Result(
        keycode=keycode,
        layout=layout + 1,
        keysym=keysym_str,
        unicode=unicode,
        active_mods=active_mods,
        consumed_mods=consumed_mods,
        level=level + 1,
        leds=leds,
    )


###############################################################################
# Pythonic interface
###############################################################################


@dataclass
class KeyLevel:
    """
    Output of a key at a specific group and level.
    """

    keycode: int
    group: int
    level: int
    keysyms: tuple[int, ...]


class ForeignKeymap:
    """
    Context manager to ensure proper handling of foreign `xkb_keymap` object.

    Intended use::

        with ForeignKeymap(xkb_base, layout="de") as keymap:
            with ForeignState(keymap) as state:
                # Use state safely here
                state.process_key_event(...)
    """

    def __init__(
        self,
        xkb_base: Path | Sequence[Path],
        rules: Optional[str] = None,
        model: Optional[str] = None,
        layout: Optional[str] = None,
        variant: Optional[str] = None,
        options: Optional[str] = None,
    ):
        self.xkb_roots = (xkb_base,) if isinstance(xkb_base, Path) else xkb_base
        self._keymap = POINTER(xkb_keymap)()  # NULL pointer
        self.rules = rules
        self.model = model
        self.layout = layout
        self.variant = variant
        self.options = options

    def __enter__(self) -> xkb_keymap_p:
        self._keymap = load_keymap(
            self.xkb_roots,
            model=self.model,
            layout=self.layout,
            variant=self.variant,
            options=self.options,
        )
        return self._keymap

    def __exit__(self, exc_type, exc_val, exc_tb):
        unref_keymap(self._keymap)

    def check(self) -> bool:
        """
        Test if keymap compiles with corresponding RMLVO config.
        """
        try:
            with self:
                return bool(self._keymap)
        except ValueError:
            return False

    def as_string(self) -> str:
        """
        Export a keymap as a string, or return an empty string on error.
        """
        try:
            with self as keymap:
                if not bool(keymap):
                    return ""
                return keymap_as_string(keymap)
        except ValueError:
            return ""

    @classmethod
    def __iter_key(
        cls, result: list[KeyLevel], keymap: xkb_keymap_p, keycode: int, data: c_void_p
    ):
        """
        Given a keymap and a keycode, iterate over its groups and levels and
        append its outputs to a list of results.
        """
        max_groups = xkb_keymap_num_layouts_for_key(keymap, keycode)
        for g in range(0, max_groups):
            max_levels = xkb_keymap_num_levels_for_key(keymap, keycode, g)
            for l in range(0, max_levels):
                keysyms = xkb_keymap_key_get_syms_by_level(keymap, keycode, g, l)
                result.append(KeyLevel(keycode, g, l, keysyms))

    @classmethod
    def get_keys_levels(cls, keymap: xkb_keymap_p) -> list[KeyLevel]:
        """
        Given a keymap, get the list of outputs for each key, group and level.
        """
        result: list[KeyLevel] = []
        xkb_keymap_key_for_each(keymap, partial(cls.__iter_key, result), c_void_p())
        return result


class State:
    """
    Convenient interface to use `xkb_state`.

    Intended use: see `ForeignKeymap`.
    """

    def __init__(self, state: xkb_state_p):
        self._state = state

    def process_key_event(self, key: str, direction: xkb_key_direction) -> Result:
        return process_key_event(self._state, key, direction)


class ForeignState:
    """
    Context manager to ensure proper handling of foreign `xkb_state` object.

    Intended use: see `ForeignKeymap`.
    """

    def __init__(self, keymap: xkb_keymap_p):
        self._keymap = keymap
        self._state = POINTER(xkb_state)()  # NULL pointer

    def __enter__(self) -> State:
        self._state = new_state(self._keymap)
        return State(self._state)

    def __exit__(self, exc_type, exc_val, exc_tb):
        unref_state(self._state)


def has_vmod_queries(xkb_base: Path) -> bool:
    """
    Check if the used xkbcommon support querying virtual modifiers
    """
    with ForeignKeymap(xkb_base) as keymap:
        mod = xkb_keymap_mod_get_index(keymap, "Alt")
        Alt = ModifierMask((1 << mod) if mod != XKB_MOD_INVALID else 0)
        assert Alt.value != 0
        with ForeignState(keymap) as state:
            # This is very indirect: if vmods are supported in queries, then
            # we expect Result.active_mods to contain rmods and relevant vmods;
            # else we will get only rmods.
            r = state.process_key_event("LALT", xkb_key_direction.XKB_KEY_DOWN)
            r = state.process_key_event("AD01", xkb_key_direction.XKB_KEY_DOWN)
            return Alt in r.active_mods



# SPDX-License-Identifier: MIT

import dataclasses
import textwrap

from rules.generator.parser import (
    MLVO,
    BlankLine,
    Comment,
    Group,
    KcCGST,
    LayoutRange,
    Rule,
    RulesFile,
    RulesSet,
    Version,
)
from rules.generator.template import Template


def test_template():
    raw = textwrap.dedent("""\
    line 1
    # # This is a comment
    # # x = 2
    line 4
    line 5 x={x} {y:#04x} {{x}}

    # for i in (0, x - 1):
    #     for j in (1, 2):
    {j}. Item {i}
       …
    #     endfor
    ---
    # endfor
    ⁂
    """)

    template = Template.parse(raw.splitlines(keepends=True))
    print(template.code)

    expected = textwrap.dedent("""\
    line 1
    line 4
    line 5 x=2 0x7b {x}

    1. Item 0
       …
    2. Item 0
       …
    ---
    1. Item 1
       …
    2. Item 1
       …
    ---
    ⁂
    """)
    got = template.render(x=2, y=123)

    assert got == expected


def test_rule():
    rule = Rule(mlvo=("a", "b"), section=r"+%l[%i]%(v[%i]):%i")
    assert rule.to_numeric_index(0) == Rule(mlvo=("a", "b"), section=r"+%l%(v)")
    assert rule.to_numeric_index(1) == Rule(mlvo=("a", "b"), section=r"+%l[1]%(v[1]):1")
    assert rule.to_numeric_index(2) == Rule(mlvo=("a", "b"), section=r"+%l[2]%(v[2]):2")


def test_rules_set_numeric_index():
    rule = Rule(mlvo=("a", "b"), section=r"+%l[%i]%(v[%i]):%i")
    rules_set = RulesSet(
        mlvo=(MLVO.Layout, MLVO.Variant),
        layout_range=LayoutRange.Any,
        kccgst=KcCGST.Symbols,
        rules=[rule],
    )
    numeric = tuple(rules_set.to_numeric_indices(4))
    assert numeric == (
        dataclasses.replace(
            rules_set, layout_range=0, rules=[Rule(mlvo=("a", "b"), section=r"+%l%(v)")]
        ),
        dataclasses.replace(
            rules_set,
            layout_range=1,
            rules=[Rule(mlvo=("a", "b"), section=r"+%l[1]%(v[1]):1")],
        ),
        dataclasses.replace(
            rules_set,
            layout_range=2,
            rules=[Rule(mlvo=("a", "b"), section=r"+%l[2]%(v[2]):2")],
        ),
        dataclasses.replace(
            rules_set,
            layout_range=3,
            rules=[Rule(mlvo=("a", "b"), section=r"+%l[3]%(v[3]):3")],
        ),
        dataclasses.replace(
            rules_set,
            layout_range=4,
            rules=[Rule(mlvo=("a", "b"), section=r"+%l[4]%(v[4]):4")],
        ),
    )


def test_rules_set_serialize():
    rule1 = Rule(mlvo=("a", "bbbbbbbbbbbbbbbbbbb"), section=r"+%l[%i]%(v[%i]):%i")
    rule2 = Rule(mlvo=("$aaaaaaaaaaaaaa", "$b"), section=r"+%l[%i]%(v[%i]):%i")
    rules_set = RulesSet(
        mlvo=(MLVO.Layout, MLVO.Variant),
        layout_range=LayoutRange.Any,
        kccgst=KcCGST.Symbols,
        rules=[rule1, rule2],
    )

    # V2
    expected = textwrap.dedent("""\
        ! layout		 variant		=	symbols
          a			 bbbbbbbbbbbbbbbbbbb	=	+%l%(v)
         $aaaaaaaaaaaaaa	$b			=	+%l%(v)

        ! layout[1]		 variant[1]		=	symbols
          a			 bbbbbbbbbbbbbbbbbbb	=	+%l[1]%(v[1]):1
         $aaaaaaaaaaaaaa	$b			=	+%l[1]%(v[1]):1

        ! layout[2]		 variant[2]		=	symbols
          a			 bbbbbbbbbbbbbbbbbbb	=	+%l[2]%(v[2]):2
         $aaaaaaaaaaaaaa	$b			=	+%l[2]%(v[2]):2

        ! layout[3]		 variant[3]		=	symbols
          a			 bbbbbbbbbbbbbbbbbbb	=	+%l[3]%(v[3]):3
         $aaaaaaaaaaaaaa	$b			=	+%l[3]%(v[3]):3

        ! layout[4]		 variant[4]		=	symbols
          a			 bbbbbbbbbbbbbbbbbbb	=	+%l[4]%(v[4]):4
         $aaaaaaaaaaaaaa	$b			=	+%l[4]%(v[4]):4
    """)
    got = "".join(rules_set.serialize(Version.V2))
    assert got == expected

    # V3
    expected = textwrap.dedent("""\
        ! layout[any]		 variant[any]		=	symbols
          a			 bbbbbbbbbbbbbbbbbbb	=	+%l[%i]%(v[%i]):%i
         $aaaaaaaaaaaaaa	$b			=	+%l[%i]%(v[%i]):%i
    """)
    got = "".join(rules_set.serialize(Version.V3))
    assert got == expected


def test_group():
    raw = "! $g = 11 2 4 1 3 "
    g = Group(name="$g", members={"1", "11", "2", "3", "4"})
    assert Group.parse(raw) == g
    assert "".join(g.serialize()) == "! $g = 1 2 3 4 11\n"


def test_rules_file():
    raw = textwrap.dedent("""\
    //
    // Some comment
    //

    // Group comment
    // on 2 lines
    ! $g1 = e1 e2 \\
            e3 // comment
    ! model layout[first] \
       = symbols

      *       a         = A/a:%i
      * * =             %l[%i]%(v[%i]):%i
    """)

    expected = (
        Comment(text="\n Some comment\n"),
        BlankLine(),
        Group(
            name="$g1",
            members={"e1", "e2", "e3"},
            component=None,
            description="Group comment\non 2 lines",
        ),
        RulesSet(
            mlvo=(MLVO.Model, MLVO.Layout),
            layout_range=LayoutRange.First,
            kccgst=KcCGST.Symbols,
            rules=[
                Rule(mlvo=("*", "a"), section="A/a:%i"),
                Rule(mlvo=("*", "*"), section="%l[%i]%(v[%i]):%i"),
            ],
        ),
    )
    got = tuple(RulesFile.parse(raw.splitlines(keepends=True)))
    assert got == expected

    expected = textwrap.dedent("""\
        //
        // Some comment
        //

        // Group comment
        // on 2 lines
        ! $g1 = e1 e2 e3

        ! model	layout[first]	=	symbols
          *	a		=	A/a:%i
          *	*		=	%l[%i]%(v[%i]):%i
    """)
    got = "".join(RulesFile.serialize(got, version=Version.V3))
    assert got == expected



#!/usr/bin/env python3
# SPDX-License-Identifier: MIT

import argparse
from pathlib import Path

from . import generate_rules, generate_symbols
from .parser import Version


def run_generate_rules(args: argparse.Namespace):
    print(
        *generate_rules(
            args.template,
            version=args.version,
            ruleset=args.ruleset,
            compat=args.compat,
            debug=args.debug,
        ),
        sep="",
        end="",
        file=args.output,
    )


def run_generate_symbols(args: argparse.Namespace):
    for symbol_file in generate_symbols(destination=args.dest):
        # Print path to stdout, for path collection in meson
        print(symbol_file.path.name)
        if args.dry_run:
            continue
        with symbol_file.path.open("wt", encoding="utf-8") as fd:
            fd.write(symbol_file.content)


parser = argparse.ArgumentParser("XKB rules generator")

subparsers = parser.add_subparsers()

rules_parser = subparsers.add_parser("rules", help="Generates rules")
rules_parser.add_argument(
    "template",
    type=argparse.FileType("rt", encoding="utf-8"),
    help="Rules template file",
)
rules_parser.add_argument(
    "--ruleset",
    type=str,
    choices=("base", "evdev"),
    default="evdev",
    help="Base rules (default: %(default)s)",
)
rules_parser.add_argument("--compat", action="store_true", help="Compatibility rules")
rules_parser.add_argument(
    "--version", type=Version.parse, default=Version.V2, help="Format version"
)
rules_parser.add_argument("--debug", action="store_true", help="Debug mode")
rules_parser.add_argument(
    "--output", type=argparse.FileType("w", encoding="UTF-8"), help="Output"
)
rules_parser.set_defaults(func=run_generate_rules)

symbols_parser = subparsers.add_parser(
    "symbols", help="Generates compatibility symbols"
)
symbols_parser.add_argument("dest", type=Path, help="Destination directory")
symbols_parser.add_argument("--dry-run", action="store_true", help="Dry run")
symbols_parser.set_defaults(func=run_generate_symbols)

args = parser.parse_args()

args.func(args)



# SPDX-License-Identifier: MIT

import itertools
import re
import string
from collections import defaultdict
from dataclasses import dataclass
from pathlib import Path
from typing import Iterable, TextIO

from .compatibility import ROOT, RULES, SYMBOLS, CompatMappings, CompatSymbolsMapping
from .options import Section
from .parser import RulesFile, Version
from .registry import Registry
from .template import Template


def generate_rules(
    template_content: Iterable[str],
    version: Version,
    ruleset: str,
    compat: bool,
    debug: bool,
) -> Iterable[str]:
    """
    Generate rules file from template
    """
    template = Template.parse(template_content)
    # “base” is used as a template for all rulesets.
    registry_files = (RULES / "base.xml", RULES / "base.extras.xml")
    registries = tuple(map(Registry.load, registry_files))

    # Options: render template with empty parameters and detect static options
    rules = template.render(
        groups=(),
        ruleset=ruleset,
        compat=compat,
        layouts_compat_mappings=[],
        variants_compat_mappings=[],
        options=defaultdict(tuple),
    )
    options = {
        s: tuple(s.get_options(xkb_root=ROOT, registries=registries, rules=rules))
        for s in Section
    }

    groups = tuple(itertools.chain.from_iterable(r.groups.values() for r in registries))

    # Compat mappings (aliases)
    compat_mappings = CompatMappings.load(
        layouts_path=RULES / "compat" / "layoutsMapping.lst" if compat else None,
        variants_path=RULES / "compat" / "variantsMapping.lst" if compat else None,
        vendors_path=RULES / "compat" / "variantsMapping-vendors.lst",
        skip_if_source_file_exists=True,
    )

    rules = template.render(
        groups=groups,
        ruleset=ruleset,
        compat=compat,
        layouts_compat_mappings=compat_mappings.layouts,
        variants_compat_mappings=compat_mappings.variants + compat_mappings.vendors,
        options=options,
    )
    return RulesFile.render(rules.splitlines(), version=version, debug=debug)


@dataclass
class SymbolsFile:
    path: Path
    content: str


SYMBOLS_TEMPLATE = string.Template("""
// Compatibility mapping
partial xkb_symbols "${alias}" {
    include "${target}"
};
""")


def generate_symbols(destination: Path) -> Iterable[SymbolsFile]:
    """
    Append xkb_symbols compat entries
    """
    mappings = CompatMappings.load(
        variants_path=RULES / "compat" / "variantsMapping.lst",
        skip_if_source_file_exists=False,
    )

    # Group by alias symbol file
    files = defaultdict(list)
    for mapping in mappings.variants:
        files[mapping.source.layout].append(mapping)

    for filename, mappings in files.items():
        src_path: Path = SYMBOLS / filename
        # Get original file content
        content = src_path.read_text(encoding="utf-8")
        # Check that there is no clash with existing sections
        new_sections = "|".join(re.escape(m.source.variant) for m in mappings)
        pattern = re.compile(rf'xkb_symbols\s+"(?P<section>{new_sections})"')
        for line_number, line in enumerate(content.splitlines(), start=1):
            # Drop comments
            line = line.split("//")[0]
            # Check for clashing section definition
            if m := pattern.search(line):
                l1 = mappings[0].source
                section = m.group("section")
                raise ValueError(
                    f'Cannot add compatibility section in symbols/{l1.layout}: "{section}" already exists at line {line_number}'
                )
        # Add compat sections
        for m in mappings:
            content += SYMBOLS_TEMPLATE.substitute(
                alias=m.source.variant, target=m.destination
            )

        yield SymbolsFile(path=destination / filename, content=content)



# SPDX-License-Identifier: MIT

from __future__ import annotations

import functools
import re
import textwrap
from dataclasses import dataclass
from pathlib import Path
from typing import ClassVar, Iterable, Self, Sequence

ROOT = Path(__file__).parent.parent.parent
RULES = ROOT / "rules"
SYMBOLS = ROOT / "symbols"
# Some checks in case we move this script
assert RULES.is_dir(), (
    f"Unexpected directory tree, expected '{RULES}/' - did this script move?"
)
assert SYMBOLS.is_dir(), (
    f"Unexpected directory tree, expected '{SYMBOLS}/' - did this script move?"
)


@functools.total_ordering
@dataclass(frozen=True, order=False)
class Layout:
    PATTERN: ClassVar[re.Pattern[str]] = re.compile(
        r"(?P<layout>[^(]+)\((?P<variant>[^)]+)\)"
    )
    DEFAULT_MODEL: ClassVar[str] = "*"
    DEFAULT_SYMBOLS_PREFIX: ClassVar[str] = "pc"

    layout: str
    variant: str
    model: str = DEFAULT_MODEL
    symbols_prefix: str = DEFAULT_SYMBOLS_PREFIX

    def __str__(self) -> str:
        if self.variant:
            return f"{self.layout}({self.variant})"
        else:
            return self.layout

    def __lt__(self, other):
        """
        Custom compare function in order to deal with missing variant.
        """
        if not isinstance(other, self.__class__):
            return NotImplemented
        elif self.model != other.model:
            if (self.model == "*") ^ (other.model == "*"):
                return other.model == "*"
            else:
                return self.model < other.model
        elif (self.layout == "*") ^ (other.layout == "*"):
            return other.layout == "*"
        elif self.layout.startswith("$") ^ other.layout.startswith("$"):
            return other.layout.startswith("$")
        elif self.layout == other.layout:
            if self.variant == other.variant:
                return False
            # Handle missing variant
            elif self.variant and (not other.variant or other.variant == "*"):
                return True
            # Handle missing variant
            elif (not self.variant or self.variant == "*") and other.variant:
                return False
            else:
                return self.variant < other.variant
        else:
            return self.layout < other.layout

    @classmethod
    def parse(
        cls,
        raw: str,
        model: str = DEFAULT_MODEL,
        symbols_prefix: str = DEFAULT_SYMBOLS_PREFIX,
    ) -> Self:
        if m := cls.PATTERN.match(raw):
            return cls(
                layout=m.group("layout"),
                variant=m.group("variant"),
                model=model,
                symbols_prefix=symbols_prefix,
            )
        else:
            return cls(
                layout=raw, variant="", model=model, symbols_prefix=symbols_prefix
            )


@dataclass(frozen=True, order=True)
class CompatSymbolsMapping:
    source: Layout
    destination: Layout

    @classmethod
    def parse(cls, raw: str, is_vendor_symbols: bool) -> Self:
        # Drop comment
        raw = raw.split("//")[0]
        parts = raw.split()

        if is_vendor_symbols:
            if len(parts) != 4:
                raise ValueError(raw)
            source = Layout.parse(parts[2], model=parts[0])
            destination = Layout.parse(parts[3], symbols_prefix=parts[1])
        else:
            match len(parts):
                case 2:
                    source = Layout.parse(parts[0])
                    destination = Layout.parse(parts[1])
                case 4:
                    source = Layout(layout=parts[0], variant=parts[1])
                    destination = Layout(layout=parts[2], variant=parts[3])
                case _:
                    raise ValueError(raw)

        return cls(source=source, destination=destination)

    @classmethod
    def parse_iter(
        cls, lines: Iterable[str], is_vendor_symbols: bool
    ) -> Iterable[Self]:
        for line in lines:
            line, *_ = line.split("//")
            if not line:
                continue
            yield cls.parse(line, is_vendor_symbols)

    @classmethod
    def parse_array(cls, raw: str, is_vendor_symbols: bool) -> Iterable[Self]:
        lines = textwrap.dedent(raw).splitlines()
        yield from cls.parse_iter(lines, is_vendor_symbols=is_vendor_symbols)


@dataclass(frozen=True)
class CompatMappings:
    layouts: Sequence[CompatSymbolsMapping]
    variants: Sequence[CompatSymbolsMapping]
    vendors: Sequence[CompatSymbolsMapping]

    @classmethod
    def load(
        cls,
        layouts_path: Path | None = None,
        variants_path: Path | None = None,
        vendors_path: Path | None = None,
        skip_if_source_file_exists: bool = True,
    ) -> Self:
        def check_symbols(layout: Layout, skip_if_source_file_exists: bool) -> bool:
            return skip_if_source_file_exists == (
                not (SYMBOLS / layout.source.layout).is_file()
            )

        if layouts_path:
            with layouts_path.open("rt", encoding="utf-8") as fd:
                layouts = sorted(
                    filter(
                        functools.partial(
                            check_symbols,
                            skip_if_source_file_exists=skip_if_source_file_exists,
                        ),
                        CompatSymbolsMapping.parse_iter(fd, is_vendor_symbols=False),
                    )
                )
        else:
            layouts = []

        if variants_path:
            with variants_path.open("rt", encoding="utf-8") as fd:
                variants = sorted(
                    filter(
                        functools.partial(
                            check_symbols,
                            skip_if_source_file_exists=skip_if_source_file_exists,
                        ),
                        CompatSymbolsMapping.parse_iter(fd, is_vendor_symbols=False),
                    )
                )
        else:
            variants = []

        if vendors_path:
            with vendors_path.open("rt", encoding="utf-8") as fd:
                vendors = sorted(
                    CompatSymbolsMapping.parse_iter(fd, is_vendor_symbols=True)
                )
        else:
            vendors = []

        return cls(layouts=layouts, variants=variants, vendors=vendors)



#!/usr/bin/env python3

"""
Query the XKB registry via xkbcommon and process the results, e.g. filter and
export (non-)Latin layouts.
"""

from __future__ import annotations

import argparse
import csv
import dataclasses
import logging
import subprocess
import sys
from collections.abc import Generator, Sequence
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Sequence

import yaml

# Add our internal xkbcommon test lib to the PATH
ROOT = Path(__file__).parent.parent
sys.path.append(str(ROOT / "tests"))

import xkbcommon

logger = logging.getLogger()
logging.basicConfig(
    stream=sys.stderr, level=logging.INFO, format="[%(levelname)s] %(message)s"
)


@dataclass
class Model:
    name: str


@dataclass(frozen=True, order=True)
class Layout:
    layout: str
    variant: str
    description: str
    extra: dict[str, Any]

    @classmethod
    def parse(cls, raw: dict[str, str]) -> Layout:
        """
        Parse YAML entry
        """
        return Layout(
            layout=raw.get("layout", ""),
            variant=raw.get("variant", ""),
            description=raw.get("description", ""),
            extra={},
        )


class Option:
    name: str


@dataclass
class Registry:
    """
    The XKB registry, i.e. a rules/*.xml file parsed.
    """

    models: tuple[Model, ...]
    layouts: tuple[Layout, ...]
    options: tuple[Option, ...]

    @classmethod
    def parse(cls, raw: dict[str, Sequence[Any]], skip_custom: bool = True) -> Registry:
        """
        Parse YAML entry
        """
        return cls(
            models=(),  # FIXME: process models
            layouts=tuple(
                l
                for l in map(Layout.parse, raw.get("layouts", ()))
                if not skip_custom or l.layout != "custom"
            ),
            options=(),  # FIXME: process options
        )

    @classmethod
    def load(
        cls, xkb_roots: Sequence[Path] | None = None, rules: str | None = None
    ) -> Registry:
        """
        Run xkbcli list and parse its YAML output.
        """
        args: tuple[str, ...] = ("xkbcli", "list", "--load-exotic")
        if xkb_roots:
            # If no xkb config root is provided, we rely on the defaults that xkbcommon
            # will pick. It depends on its built-in defaults and on the environment.
            args += ("--skip-default-paths",)
            args += tuple(map(str, xkb_roots))
        if rules:
            # If no rules set is provided, we rely on the default one that xkbcommon
            # will pick. It depends on its built-in default and on the environment.
            args += ("--ruleset", rules)
        logger.info(f"Running: {' '.join(args)}")
        p = subprocess.run(args, encoding="utf-8", capture_output=True)
        raw = yaml.safe_load(p.stdout)
        return cls.parse(raw)


class Csv(csv.unix_dialect):
    """
    CSV dialect used to export results.
    """

    quoting = csv.QUOTE_NONE


LATIN_LETTERS = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
"""
Required upper case characters for a layout to be considered Latin.
"""


def filter_latin_layouts(
    xkb_roots: Sequence[Path],
    registry: Registry,
    latin: bool,
    rules: str | None = None,
    debug: bool = False,
) -> Generator[Layout, None, None]:
    """
    Given a registry, filter all its layouts that are (non-)Latin by checking
    that each required characters are accessible at some key, group and level.
    """
    latin_letters = frozenset(LATIN_LETTERS + LATIN_LETTERS.lower())
    latin_keysyms: dict[int, str] = {
        xkbcommon.xkb_keysym_from_name(c): c for c in latin_letters
    }
    for layout in registry.layouts:
        try:
            with xkbcommon.ForeignKeymap(
                xkb_base=xkb_roots,
                rules=rules,
                layout=layout.layout,
                variant=layout.variant,
            ) as keymap:
                found: set[str] = set()
                r: xkbcommon.KeyLevel
                for r in xkbcommon.ForeignKeymap.get_keys_levels(keymap):
                    for k in r.keysyms:
                        if (c := latin_keysyms.get(k)) is not None:
                            found.add(c)
        except ValueError as err:
            # Log error message and skip
            logger.error(err)
            continue
        missing = latin_letters.difference(found)
        if latin ^ bool(missing):
            if debug:
                # Add missing characters for debugging
                extra = dict(layout.extra)
                extra["missing"] = missing_str = "".join(sorted(missing))
                yield dataclasses.replace(layout, extra=extra)
            else:
                yield layout
        almost_latin = len(missing) / len(latin_letters) <= 0.10
        if debug and missing and almost_latin:
            logger.debug(
                f"Almost a Latin layout: {layout}; missing: {missing_str} ({len(missing)})"
            )


def process_layouts(
    xkb_roots: Sequence[Path], registry: Registry, args: argparse.Namespace
):
    """
    Process layouts from a given registry, depending on the CLI arguments.
    """
    debug: bool = args.debug
    if args.latin or args.non_latin:
        # Filter (non-)Latin layouts
        layouts = tuple(
            filter_latin_layouts(
                xkb_roots,
                registry,
                latin=args.latin,
                rules=args.rules,
                debug=args.debug,
            )
        )
    else:
        # Get all layouts
        layouts = registry.layouts
    if args.csv:
        # Output as CSV
        path: Path = args.csv
        with path.open("wt", encoding="utf-8", newline="") as fd:
            writer = csv.writer(fd, dialect=csv.unix_dialect if debug else Csv)
            fields: tuple[str, ...] = ("Layout", "Variant")
            if debug:
                fields += (
                    "Description",
                    "Missing Latin characters",
                )
            writer.writerow(fields)

            def get_fields(layout) -> Generator[str, None, None]:
                yield layout.layout
                yield layout.variant
                if debug:
                    yield layout.description
                    yield layout.extra.get("missing", "")

            for layout in sorted(layouts):
                writer.writerow(get_fields(layout))
    else:
        # Output as Python representation, for debugging
        for layout in layouts:
            print(layout)


def parse_cli_args() -> argparse.Namespace:
    """
    Create CLI parser and parse corresponding arguments.
    """
    parser = argparse.ArgumentParser()
    parser.add_argument("--xkb-root", type=Path, action="append", required=True)
    parser.add_argument("--debug", action="store_true")
    parser.add_argument("--rules", help="Rules set to use")
    subparsers = parser.add_subparsers(required=True)
    subparser = subparsers.add_parser("layouts", help="List layouts")
    subparser.set_defaults(run=process_layouts)
    subparser.add_argument("--csv", type=Path)
    group = subparser.add_mutually_exclusive_group()
    group.add_argument("--latin", action="store_true", help="List only Latin layouts")
    group.add_argument(
        "--non-latin", action="store_true", help="List only non-Latin layouts"
    )
    return parser.parse_args()


if __name__ == "__main__":
    args = parse_cli_args()
    if args.debug:
        logger.setLevel(logging.DEBUG)
    xkb_roots: list[Path] = args.xkb_root
    rules: str | None = args.rules
    registry = Registry.load(xkb_roots, rules)
    args.run(xkb_roots, registry, args)



#!/usr/bin/env python3
# SPDX-License-Identifier: MIT

"""
Minimal and unsafe template engine based on `exec()`

Only a very limited subset of Python is allowed in the template.
However, this template engine is still considered unsafe, due to
the intrinsic unsafe nature of `exec()`. Nevertheless, this is deemed
sufficient because because both the template and the input data are
version controlled, and a successful attack would require malicious
content that would be conspicuous in any routine code review.
"""

from __future__ import annotations

import argparse
import ast
import re
import textwrap
from dataclasses import dataclass
from typing import Any, ClassVar, Iterable, Self


@dataclass
class Template:
    code: str

    INDENT_PATTERN: ClassVar[re.Pattern[str]] = re.compile(r"^( +)")
    KEYWORD_PATTERN: ClassVar[re.Pattern[str]] = re.compile(
        r"^\s*(end(?:if|for|while|with|def|class)|"
        r"if|elif|else|for|while|with|def|class)"
    )
    RENDER: ClassVar[str] = "render"
    SAFE_NODES: ClassVar[tuple[ast.AST, ...]] = (
        ast.Module,
        ast.If,
        ast.IfExp,
        ast.Expr,
        ast.Expression,
        ast.Constant,
        ast.Name,
        ast.Load,
        ast.UnaryOp,
        ast.BoolOp,
        ast.BinOp,
        ast.Add,
        ast.Sub,
        ast.Mult,
        ast.Div,
        ast.FloorDiv,
        ast.Mod,
        ast.Pow,
        ast.LShift,
        ast.RShift,
        ast.BitOr,
        ast.BitXor,
        ast.BitAnd,
        ast.MatMult,
        ast.Compare,
        ast.Eq,
        ast.NotEq,
        ast.Lt,
        ast.LtE,
        ast.Gt,
        ast.GtE,
        ast.Is,
        ast.IsNot,
        ast.In,
        ast.NotIn,
        ast.JoinedStr,
        ast.FormattedValue,
        ast.List,
        ast.Tuple,
        ast.Dict,
        ast.Attribute,
        ast.Yield,
        ast.For,
        ast.Subscript,
        ast.Call,
        ast.Or,
        ast.And,
        ast.Not,
        ast.Store,
        ast.Continue,
    )

    # Unused for now. Requires to add the functions explicitly in the "__builtins__" globals
    SAFE_FUNCTIONS: ClassVar[tuple[str, ...]] = ("range",)

    @classmethod
    def parse(cls, raw: Iterable[str]) -> Self:
        code = "\n".join(cls._parse(raw))

        # Guard allowed Python tokens as a minimal security

        try:
            _ast = ast.parse(code)
        except SyntaxError as exc:
            raise exc

        for node in ast.walk(_ast):
            if type(node) not in cls.SAFE_NODES:
                raise PermissionError(f"Unsupported node: {node}")
            elif isinstance(node, ast.Attribute):
                if node.attr.startswith("_"):
                    raise PermissionError(
                        f"Access to private attribute '{node.attr}' is forbidden."
                    )
            elif isinstance(node, ast.JoinedStr):
                # Guard against malicious values in the registry
                for v in node.values:
                    if not cls._is_simple(v):
                        raise PermissionError(f"Invalid simple f-string: {node}")
            elif isinstance(node, ast.Subscript):
                # Only allow `xxx[yyy]` where yyy is a constant that does not start with "_"
                match node.slice:
                    case ast.Constant(value=str(s)) if s.startswith("_"):
                        raise PermissionError(
                            f"Access to private subscript '{s}' is forbidden."
                        )
                    case ast.Constant():
                        pass  # int, non-private string: fine
                    case _:
                        raise PermissionError(f"Non-constant subscript is forbidden.")
            elif isinstance(node, ast.Call):
                # Only allow functions that are class methods
                # Only allow `.items()`
                if (
                    not isinstance(node.func, ast.Attribute)
                    or node.func.attr != "items"
                    or node.args
                    or node.keywords
                ) and (
                    not isinstance(node.func, ast.Name)
                    or node.func.id not in cls.SAFE_FUNCTIONS
                ):
                    raise PermissionError(
                        f"Access to function '{node.func}' is forbidden."
                    )

        code = f"def {cls.RENDER}():\n" + textwrap.indent(code, "    ")

        return cls(code)

    @classmethod
    def _parse(cls, raw: Iterable[str]) -> Iterable[str]:
        indent = ""
        for line in raw:
            if line.startswith("#"):
                line = line[2:]
                if not line:
                    continue
                if line.lstrip().startswith("#"):
                    # Skip comment
                    continue
                if m := cls.INDENT_PATTERN.match(line):
                    indent = m.group(1)
                else:
                    indent = ""
                if m := cls.KEYWORD_PATTERN.match(line):
                    if m.group(1).startswith("end"):
                        continue
                    indent += "    "
                yield line
            else:
                yield indent + f"yield f{repr(line)}"

    @classmethod
    def _is_simple(cls, node: ast.JoinedStr) -> bool:
        match node:
            case ast.Constant(value=value):
                return True
            case ast.FormattedValue(value=value, conversion=-1):
                return cls._is_simple_expr(value)
            case _:
                return False

    @classmethod
    def _is_simple_expr(cls, node: ast.AST) -> bool:
        """Only allow plain names and attribute access chains"""
        match node:
            case ast.Name():
                return not node.id.startswith("_")
            case ast.Attribute(value=inner, attr=attr_):
                return cls._is_simple_expr(inner) and not attr_.startswith("_")
            case _:
                return False

    def render(self, **kwargs) -> str:
        context: dict[str, Any] = dict(kwargs)
        context["__builtins__"] = {"range": range}
        exec(self.code, context)
        return "".join(context[self.RENDER]())


def run_render(args: argparse.Namespace):
    template = Template.parse(args.path)
    print(template.render())


def run_tests(args: argparse.Namespace):
    test_basic()


def test_basic():
    template = Template.parse(
        """\
# if x < 1:
    a
# elif x < 5:
    b
# else:
    c
# endif
""".splitlines()
    )

    output = template.render(x=3)
    assert output == "    b", output

    try:
        template = Template.parse(
            """
# [].__class__.__base__.__subclasses__()[140].__init__.__globals__['sys'].modules['os']
""".splitlines()
        )
    except PermissionError as e:
        ...
    else:
        raise Exception("Expection not raised")

    try:
        template = Template.parse(
            """
# (x for x in []).gi_frame.f_builtins['__import__']('os').system('whoami')
""".splitlines()
        )
    except PermissionError as e:
        ...
    else:
        raise Exception("Expection not raised")

    try:
        template = Template.parse("""! {x.__class__} = x""".splitlines())
    except PermissionError as e:
        ...
    else:
        raise Exception("Expection not raised")

    try:
        template = Template.parse("""{ { ().__class__ : "safe" } }""".splitlines())
    except PermissionError as e:
        ...
    else:
        raise Exception("Expection not raised")


if __name__ == "__main__":
    parser = argparse.ArgumentParser("Template rendering")
    subparsers = parser.add_subparsers()

    parser_render = subparsers.add_parser("render", help="Render template")
    parser_render.add_argument("path", type=argparse.FileType("rt", encoding="utf-8"))
    parser_render.set_defaults(run=run_render)

    parser_test = subparsers.add_parser("test", help="Run tests")
    parser_test.set_defaults(run=run_tests)

    args = parser.parse_args()
    args.run(args)



# SPDX-License-Identifier: MIT

from __future__ import annotations

import argparse
import sys
import xml.etree.ElementTree as ET
from collections.abc import Iterable
from dataclasses import dataclass
from enum import StrEnum, unique
from pathlib import Path

from .parser import MLVO, RulesFile, RulesSet
from .registry import Registry


def error(msg):
    print(f"ERROR: {msg}", file=sys.stderr)
    print("Aborting now")
    sys.exit(1)


@unique
class Section(StrEnum):
    """
    XKB sections.
    Name correspond to the header (`xkb_XXX`), value to the subdir/rules header.
    """

    keycodes = "keycodes"
    compatibility = "compat"
    geometry = "geometry"
    symbols = "symbols"
    types = "types"

    @classmethod
    def parse(cls, raw: str) -> Section:
        # Note: in order to display a nice message, argparse requires the error
        # to be one of: ArgumentTypeError, TypeError, or ValueError
        # See: https://docs.python.org/3/library/argparse.html#type
        try:
            return cls[raw]
        except KeyError:
            raise ValueError(raw)

    def get_options(
        self, xkb_root: Path, registries: Iterable[Registry], rules: str
    ) -> Iterable[tuple[str, str]]:
        all_options = (Option(opt) for r in registries for opt in r.options)

        skip = frozenset(self.find_options_to_skip(rules))

        directives = (
            o.resolve(xkb_root)
            for o in sorted(all_options)
            if o.name not in skip and not o.name.startswith("custom:")
        )

        def check_and_map(directive: DirectiveSet) -> Directive:
            assert not directive.is_empty, (
                f"Option {directive.option} does not resolve to any section"
            )

            return getattr(directive, self.name)

        filtered = filter(
            lambda y: y is not None,
            map(check_and_map, directives),
        )

        for d in filtered:
            assert d is not None
            yield (d.name, d)

        if self is Section.types:
            yield ("custom:types", "custom")

    @classmethod
    def find_options_to_skip(cls, rules: str) -> Iterable[str]:
        """
        Find options to skip

        Theses are the “option” rules defined explicitly in partial rules files *.part
        """
        for rs in RulesFile.parse(rules.splitlines()):
            if not isinstance(rs, RulesSet) or MLVO.Option not in rs.mlvo:
                # Skip comments/groups/includes and rulesets without options
                continue
            idx = rs.mlvo.index(MLVO.Option)
            for r in rs.rules:
                yield r.mlvo[idx]


@dataclass
class Directive:
    option: Option
    filename: str
    section: str

    @property
    def name(self) -> str:
        return self.option.name

    def __str__(self) -> str:
        if self.section:
            return f"{self.filename}({self.section})"
        else:
            return self.filename


@dataclass
class DirectiveSet:
    option: Option
    keycodes: Directive | None
    compatibility: Directive | None
    geometry: Directive | None
    symbols: Directive | None
    types: Directive | None

    @property
    def is_empty(self) -> bool:
        return all(
            x is None
            for x in (
                self.keycodes,
                self.compatibility,
                self.geometry,
                self.symbols,
                self.types,
            )
        )


@dataclass
class Option:
    """
    Wrapper around a single option -> KcCGST rules file entry. Has the properties
    name and directive where the directive consists of the XKB symbols file name
    and corresponding section, usually composed in the rules file as:
        name = +directive
    """

    name: str

    def __lt__(self, other) -> bool:
        return self.name < other.name

    @property
    def directive(self) -> Directive:
        f, s = self.name.split(":")
        return Directive(self, f, s)

    def resolve(self, xkb_root: Path) -> DirectiveSet:
        directives: dict[Section, Directive | None] = {s: None for s in Section}
        directive = self.directive
        filename, section_name = directive.filename, directive.section
        for section in Section:
            subdir = xkb_root / section
            if not (subdir / filename).exists():
                # Some of our foo:bar entries map to a baz_vndr/foo file
                for vndr in subdir.glob("*_vndr"):
                    vndr_path = vndr / filename
                    if vndr_path.exists():
                        filename = vndr_path.relative_to(subdir).as_posix()
                        break
                else:
                    continue

            if (subdir / filename).is_symlink():
                resolved_filename = (subdir / filename).resolve().name
                assert (subdir / filename).exists()
            else:
                resolved_filename = filename

            # Now check if the target file actually has that section
            f = subdir / resolved_filename
            with f.open("rt", encoding="utf-8") as fd:
                section_header = f'xkb_{section.name} "{section_name}"'
                if any(section_header in line for line in fd):
                    directives[section] = Directive(
                        self, resolved_filename, section_name
                    )

        return DirectiveSet(
            option=self,
            keycodes=directives[Section.keycodes],
            compatibility=directives[Section.compatibility],
            geometry=directives[Section.geometry],
            symbols=directives[Section.symbols],
            types=directives[Section.types],
        )



# SPDX-License-Identifier: MIT

from __future__ import annotations

import sys
import xml.etree.ElementTree as ET
from collections.abc import Iterable
from dataclasses import dataclass
from enum import unique
from pathlib import Path
from typing import Self

from .parser import Group


@dataclass(frozen=True)
class Registry:
    path: Path
    groups: dict[str, Group]
    options: tuple[str]

    @classmethod
    def load(cls, path: Path) -> Self:
        options = tuple(cls.load_options(path))

        return cls(
            path=path,
            groups={},  # TODO
            options=options,
        )

    @classmethod
    def load_options(cls, path: Path) -> Iterable[str]:
        """
        Yields all Options from the given XML file
        """
        tree = ET.parse(path)
        root = tree.getroot()

        for option in root.iter("option"):
            yield cls.fetch_name(option)

    @classmethod
    def fetch_subelement(cls, parent, name):
        sub_element = parent.findall(name)
        if sub_element is not None and len(sub_element) == 1:
            return sub_element[0]
        return None

    @classmethod
    def fetch_text(cls, parent, name) -> str | None:
        sub_element = cls.fetch_subelement(parent, name)
        if sub_element is None:
            return None
        return sub_element.text

    @classmethod
    def fetch_name(cls, elem) -> str:
        try:
            ci_element = (
                elem
                if elem.tag == "configItem"
                else cls.fetch_subelement(elem, "configItem")
            )
            name = cls.fetch_text(ci_element, "name")
            assert name is not None
            return name
        except AssertionError as e:
            endl = "\n"  # f{} cannot contain backslashes
            e.args = (
                f"\nFor element {ET.tostring(elem).decode('utf-8')}\n{endl.join(e.args)}",
            )
            raise e
